uniform float time;
uniform float texScaleS;
uniform float texScaleT;

uniform sampler2D waveMap;
uniform sampler2D reflectionMap;

//uniform vec4 deepColor;
//uniform vec4 shallowColor;
//uniform float reflectionAmount;
//uniform float waterAmount;
//uniform float fresnelPower;
//uniform float fresnelBias;
uniform float hdrMultiplier;
//uniform float distFactor;
//uniform float bumpFactor;

uniform sampler2D hmapTex; //sampler2DRect
uniform float hmapMin;
uniform float hmapMax;
uniform int hmapW;

uniform int fogEnabled;

uniform int shadowed;
uniform sampler2DShadow shadowMap;
uniform int shadowSize;
uniform vec4 shadowColor;

varying vec3 eyeVector;
varying vec3 lightVector;
varying float fogFragCoord;

float iglobaltime=time;

// added 'real' water height (hits terrain) and waterheight is visible on shadow
float coast2water_fadedepth  = 0.57;
float large_waveheight       = 0.50; // change to adjust the "heavy" waves
float large_wavesize         = 2.;   // factor to adjust the large wave size
float small_waveheight       = .5;   // change to adjust the small random waves
float small_wavesize         = .5;   // factor to ajust the small wave size
float water_softlight_fact   = 170.;  // range [1..200] (should be << smaller than glossy-fact)
float water_glossylight_fact = 120.;  //120.; // range [1..200]
float particle_amount        = 30.;  //70.;
vec3  watercolor             = vec3(0.1, 0.24, 0.24); // 'transparent' low-water color (RGB)
vec3  watercolor2            = vec3(0, 0.025, 0.05); // deep-water color (RGB, should be darker than the low-water color)
vec3  water_specularcolor    = vec3(0.4, 0.4, 0.4);    // specular Color (RGB) of the water-highlights
vec3  light                  = vec3(lightVector.x, lightVector.y, lightVector.z);

// calculate random value
float hash(float n)
{
   return fract(sin(n)*43758.5453123);
}

// 2d noise function
float noise1(in vec2 x)
{
   vec2 p=floor(x);
   vec2 f=smoothstep(0.0, 1.0, fract(x));
   float n=p.x+p.y*57.0;
   return mix(mix(hash(n+0.0), hash(n+1.0), f.x), mix(hash(n+57.0), hash(n+58.0), f.x), f.y);
}

float noise(vec2 p)
{
   return texture2D(waveMap, p*vec2(1./256.)).x;
}

float height_map(vec2 p)
{
   vec4 v=texture2D(hmapTex, p); //texture2DRect
   float h=hmapMin+v.w*(hmapMax-hmapMin);
   return h;
}

vec3 terrain_map(vec2 p)
{
   /*return vec3(0.7, .55, .4)+texture2D(waveMap, p*2.).rgb*.5; // test-terrain is simply 'sandstone'/**/
   return vec3(0.175, .1375, .1)+texture2D(waveMap, p*8.).rgb*.25; //return vec3(0.0);
}

const mat2 m=mat2(0.72, -1.60,  1.60,  0.72);

float water_map(vec2 p, float height)
{
   vec2 p2=p*large_wavesize;
   vec2 shift1=0.001*vec2(iglobaltime*160.0*2.0, iglobaltime*120.0*2.0);
   vec2 shift2=0.001*vec2(iglobaltime*190.0*2.0, -iglobaltime*130.0*2.0);

   // coarse crossing 'ocean' waves...
   float f=0.6000*noise(p);
   f+=0.2500*noise(p*m);
   f+=0.1666*noise(p*m*m);
   float wave=sin(p2.x*0.622+p2.y*0.622+shift2.x*4.269)*large_waveheight*f*height*height ;

   p*=small_wavesize;
   f=0.;
   float amp=1.0, s=.5;
   for (int i=0; i<9; i++)
   { p=m*p*.947; f -= amp*abs(sin((noise(p+shift1*s)-.5)*2.)); amp=amp*.59; s*=-1.329; }

   return wave+f*small_waveheight;
}

float nautic(vec2 p)
{
   /**/
   p*=18.;
   float f=0.;
   float amp=1.0, s=.5;
   for (int i=0; i<3; i++)
   { p=m*p*1.2; f+=amp*abs(smoothstep(0., 1., noise(p+iglobaltime*s))-.5); amp=amp*.5; s*=-1.227; }
   return pow(1.-f, 5.);
   /**/
   //return 1.0;
}

float particles(vec2 p)
{
   p*=200.;
   float f=0.;
   float amp=1.0, s=1.5;
   for (int i=0; i<3; i++)
   { p=m*p*1.2; f+=amp*noise(p+iglobaltime*s); amp=amp*.5; s*=-1.227; }
   return pow(f*.35, 7.)*particle_amount;
}

float test_shadow(vec2 xy, float height)
{
   float sh=1.0;
   if(shadowed==1)
   {
      float of=1.0/shadowSize;
      sh=0.0;
      sh+=shadow2DProj(shadowMap, gl_TexCoord[3]).r<0.999?0.0:1.0;
      sh+=shadow2DProj(shadowMap, gl_TexCoord[3]+vec4(of, of, 0.0, 0.0)).r<0.999?0.0:1.0;
      sh+=shadow2DProj(shadowMap, gl_TexCoord[3]-vec4(of, of, 0.0, 0.0)).r<0.999?0.0:1.0;
      sh=sh/3.0;
      sh=sh+shadowColor.x*0.7*(1.0-sh);
   }
   return sh;
}

vec3 CalcTerrain(vec2 uv, float height)
{
   vec3 col=terrain_map(uv);
   float h1=height_map(uv-vec2(0., 0.01));
   float h2=height_map(uv+vec2(0., 0.01));
   float h3=height_map(uv-vec2(0.01, 0.));
   float h4=height_map(uv+vec2(0.01, 0.));
   vec3 norm=normalize(vec3(h3-h4, h1-h2, 1.));
   vec3 r0=vec3(uv, height);
   vec3 rd=normalize(light-r0);
   float grad=dot(norm, rd);
   col*=grad+pow(grad, 8.);
   float terrainshade=test_shadow(uv, height);
   col=mix(col*.25, col, terrainshade);
   return col;
}

void mainImage(out vec4 fragColor, in vec2 fragCoord)
{
   //vec2 uv=(fragCoord.xy/iResolution.xy-vec2(-0.12, +0.25));
   vec2 uv=fragCoord.xy;

   float WATER_LEVEL=gl_TexCoord[1].z;//1.5/*0.94*/; // Water level (range: 0.0-2.0)
   float deepwater_fadedepth=0.5+coast2water_fadedepth;

   float height=height_map(uv);
   vec3  rescolor;
   float resalpha;

   float waveheight=clamp(WATER_LEVEL*3.-1.5, 0., 1.);
   float level=WATER_LEVEL+.2*water_map(uv*15.+vec2(iglobaltime*.1), waveheight);
   if (height>level)
   {
      /*
      rescolor=CalcTerrain(uv, height);
      */
      fragColor=vec4(0.0);
      return;
   }
   if (height<=level)
   {
      vec2 dif=vec2(.0, .01);
      vec2 pos=uv*15./texScaleS*(hmapW/128.0)+vec2(iglobaltime*.01); //texScaleS = 0.105
      float h1=water_map(pos-dif,waveheight);
      float h2=water_map(pos+dif,waveheight);
      float h3=water_map(pos-dif.yx,waveheight);
      float h4=water_map(pos+dif.yx,waveheight);
      vec3 normwater=normalize(vec3(h3-h4, h1-h2, 0.15)); // norm-vector of the 'bumpy' water-plane
      vec2 wavejitter=normwater.xy*texScaleT*(level-height); //texScaleT = [0.1..0.0001]
      uv+=wavejitter;

      rescolor=CalcTerrain(uv, height);

      float coastfade=clamp((level-height)/coast2water_fadedepth, 0., 1.);
      float coastfade2=clamp((level-height)/deepwater_fadedepth, 0., 1.);
      float intensity=rescolor.r*.2126+rescolor.g*.7152+rescolor.b*.0722;
      watercolor=mix(watercolor*intensity, watercolor2, smoothstep(0., 1., coastfade2)); //coastfade

      //float fdist=distFactor/length(eyeVector);
      //float facing=1.0+eyeVector.y*fdist*(0.1-wavejitter.y*bumpFactor);
      //float fresnel=clamp(fresnelBias+pow(facing, fresnelPower), 0.0, 1.0);
      //wavejitter.y=abs(wavejitter.y);
      vec3 refuv=gl_TexCoord[0].xyw;
      refuv.xy-=wavejitter*refuv.z;
      vec4 refcol=texture2DProj(reflectionMap, refuv);
      refcol.xyz*=(refcol.x+refcol.y+refcol.z)*hdrMultiplier;
      //vec4 watcol=mix(shallowColor, deepColor, facing)*waterAmount;
      //refcol=mix(watcol, refcol, fresnel)*reflectionAmount;
      watercolor=watercolor+refcol.xyz; //watercolor=mix(refcol.xyz, watercolor, 0.5);

      vec3 r0=vec3(uv, WATER_LEVEL);
      vec3 rd=normalize(light-r0); // ray-direction to the light from water-position
      float grad=dot(normwater, rd); // dot-product of norm-vector and light-direction
      float specular=pow(grad, water_softlight_fact);  // used for soft highlights
      float specular2=pow(grad, water_glossylight_fact); // used for glossy highlights
      float gradpos=dot(vec3(0., 0., 1.), rd);
      float specular1=smoothstep(0., 1., pow(gradpos, 5.));  // used for diffusity (some darker corona around light's specular reflections...)
      float watershade=test_shadow(uv, level);
      watercolor*=2.2+watershade;
      watercolor+=(.2+.8*watershade)*((grad-1.0)*.5+specular)*.25;
      watercolor/=(1.+specular1*1.25);
      watercolor+=watershade*specular2*water_specularcolor;
      watercolor+=watershade*coastfade*(1.-coastfade2)*(vec3(.5, .6, .7)*nautic(uv)+vec3(1., 1., 1.)*particles(uv));

      rescolor=mix(rescolor, watercolor, coastfade);
      resalpha=coastfade2; //coastfade
   }
   fragColor=vec4(rescolor, resalpha*1.0);
   //fragColor=vec4(watercolor, resalpha);
   //fragColor=vec4(rescolor, 1.0);
}

vec4 fogmap(vec4 fragcolor)
{
   if (fogEnabled>0)
   {
      float f=1.0;
      if (fogEnabled==1)
      f=clamp((gl_Fog.end-fogFragCoord)*gl_Fog.scale, 0.0, 1.0);
      else
      if (fogEnabled==2)
      f=clamp(exp2(-gl_Fog.density*fogFragCoord*1.442695), 0.0, 1.0);
      else
      if (fogEnabled==3)
      f=clamp(exp2(-pow(gl_Fog.density*fogFragCoord, 2.0)*1.442695), 0.0, 1.0);
      fragcolor.xyz=mix(gl_Fog.color.xyz, fragcolor.xyz, f).xyz;
   }
   return fragcolor;
}

void main()
{
   vec4 fragcolor;
   vec2 fragcoord=gl_TexCoord[1].xy;
   mainImage(fragcolor, fragcoord);
   fragcolor=fogmap(fragcolor);
   gl_FragColor=fragcolor; //vec4(1.0)
}